_ADAPTERS = {}

def register_domain_converter(
    source_domain: 'FrameBufferDomain',
    destination_domain: 'FrameBufferDomain',
    func
):
    """
    Register a `FrameBufferDomain` converter to go
    transform a `FrameBuffer` the `source_domain`
    domain to the `destination_domain` domain.
    """
    _ADAPTERS[(source_domain, destination_domain)] = func

def frame_buffer_to_domain(
    frame_buffer: 'FrameBuffer',
    target_domain: 'FrameBufferDomain',
    **kwargs
):
    """
    Convert the `frame_buffer` provided to the
    `target_domain` given, if possible.
    """
    if frame_buffer.domain == target_domain:
        return frame_buffer

    key = (frame_buffer.domain, target_domain)
    if key not in _ADAPTERS:
        raise RuntimeError(f'No adapter {key}')

    return _ADAPTERS[key](frame_buffer, **kwargs)



from dataclasses import dataclass

        
@dataclass
class RenderConfig:
    """
    The configuration of the rendering process,
    which includes information such as:
    - The `size` we want for the output
    - The `fps` we want for the output
    - If we want to use GPU or not
    """

    def __init__(
        self,
        size:  tuple[int, int],
        fps: float,
        do_use_gpu: bool
    ):
        self.size: tuple[int, int] = size
        """
        The size that will be used for the output of the
        rendering process.
        """
        self.fps: float = fps
        """
        The fps for the output of the rendering process.
        """
        self.do_use_gpu: bool = do_use_gpu
        """
        Flag to indicate if processing the rendering with
        GPU (if `True`) or with CPU (if `False`).
        """

    def deactivate_gpu(
        self
    ) -> 'RenderConfig':
        """
        Deactivate the GPU by setting the `do_use_gpu`
        parameter to `False`, if needed.
        """
        self.do_use_gpu = False

        return self

    def activate_gpu(
        self
    ) -> 'RenderConfig':
        """
        Activate the GPU by setting the `do_use_gpu`
        parameter to `True`, if needed.
        """
        self.do_use_gpu = True

        return self


from yta_editor_common.frame.core import FrameBuffer
from yta_editor_common.frame.core.domain import FrameBufferDomain
from yta_editor_common.frame.registry import register_domain_converter


# TODO: Add the 'requires_dependency' for 'moderngl' (?)
# TODO: Add the 'requires_dependency' for 'numpy' (?)
def cpu_to_gpu(
    frame: FrameBuffer,
    moderngl_context: 'moderngl.Context'
) -> FrameBuffer:
    """
    Method to convert from CPU (`numpy`) to GPU
    (`moderngl.Texture`).
    """
    from yta_editor_common.frame.helper import FrameHelper

    h, w = frame.payload.shape[:2]
    format = frame.format
    dtype = frame.dtype
    is_normalized = frame.is_normalized
    domain = FrameBufferDomain.GPU
    payload = FrameHelper.ndarray_to_texture(
        frame = frame.payload,
        moderngl_context = moderngl_context,
        do_include_alpha = True
    )

    return FrameBuffer(
        size = (w, h),
        format = format,
        dtype = dtype,
        is_normalized = is_normalized,
        domain = domain,
        payload = payload
    )

    import moderngl

    arr = frame.payload
    h, w = arr.shape[:2]

    tex = moderngl_context.texture(
        (w, h),
        components = 4,
        data = arr.tobytes()
    )

    return FrameBuffer(
        size = (w, h),
        format = frame.format,
        dtype = frame.dtype,
        is_normalized = frame.is_normalized,
        domain = FrameBufferDomain.GPU,
        payload = tex
    )

register_domain_converter(FrameBufferDomain.CPU, FrameBufferDomain.GPU, cpu_to_gpu)



from yta_validation.parameter import ParameterValidator


class RenderResources:
    """
    The resources handler of the rendering process.
    Here you have a list with the funcionalities you
    can interact with:
    - Acquiring and releasing render targets by using
    the `.render_target` attribute.
    """

    def __init__(
        self,
        opengl_context: 'moderngl.Context',
        render_target_provider: 'RenderTargetProvider',
        texture_provider: 'TexturesProvider'
    ):
        ParameterValidator.validate_mandatory_instance_of('opengl_context', opengl_context, 'Context')
        ParameterValidator.validate_mandatory_instance_of('render_target_provider', render_target_provider, 'RenderTargetProvider')
        ParameterValidator.validate_mandatory_instance_of('texture_provider', texture_provider, 'TextureProvider')

        self.opengl_context: 'moderngl.Context' = opengl_context
        """
        The moderngl context we need to operate with GPU.
        """
        self._render_target_provider: 'RenderTargetProvider' = render_target_provider
        """
        *For internal use only*

        The render target provider, able to handle the
        pool and acquire and release `RenderTarget`
        instances on demand.
        """
        self._texture_provider: 'TextureProvider' = texture_provider
        """
        *For internal use only*

        The texture provider, able to handle the pool and
        acquire and release `moderngl.Texture` instances
        on demand.
        """

        """
        TODO: Maybe having a pool for the textures, that
        is used by the RenderTarget provider is also an
        interesting thing...
        """

        """
        Exposed API below
        """
        self.render_target: '_RenderTargetAccessor' = _RenderTargetAccessor(self._render_target_provider)
        """
        The functionality related to the `RenderTarget`
        instances in one click, by using an instance of its
        provider class.
        """
        self.texture: '_TextureAccessor' = _TextureAccessor(self._texture_provider)
        """
        The functionality related to the `moderngl.Texture`
        pool handling in one click, by using an instance of
        the texture provider class.
        """
        

"""
API exposer wrappers below
"""
class _RenderTargetAccessor:
    """
    *For internal use only*

    Wrapper to simplify the access to the functionality
    that the `RenderTargetProvider` is providing to us
    and limitate the part of its API we expose.
    """

    def __init__(
        self,
        render_target_provider: 'RenderTargetProvider'
    ):
        self._render_target_provider: 'RenderTargetProvider' = render_target_provider
        """
        *For internal use only*

        The `RenderTargetProvider` instance that allows us
        to have the functionality.
        """

    def acquire(
        self,
        width: int = 1920,
        height: int = 1080,
        number_of_components: int = 4,
        dtype: str = 'f1'
    ) -> 'RenderTarget':
        """
        Get a `RenderTarget` instance with the given `width`,
        `height`, `number_of_components` and `dtype` from the
        pool, that will mark it as in use.
        """
        return self._render_target_provider.acquire(
            width = width,
            height = height,
            number_of_components = number_of_components,
            dtype = dtype
        )
    
    def acquire_like(
        self,
        render_target: 'RenderTarget'
    ) -> 'RenderTarget':
        """
        Get the `RenderTarget` instance with the same
        properties than the `render_target` provided as
        parameter.
        """
        return self._render_target_provider.acquire_like(
            render_target = render_target
        )

    def release(
        self,
        render_target: 'RenderTarget'
    ) -> None:
        """
        Release the `render_target` provided, making it
        available again through the pool and marking it
        as free.
        """
        self._render_target_provider.release(render_target)

class _TextureAccessor:
    """
    *For internal use only*

    Wrapper to simplify the access to the functionality
    that the `TextureProvider` is providing to us and
    limitate the part of its API we expose.
    """

    def __init__(
        self,
        texture_provider: 'TextureProvider'
    ):
        self._texture_provider: 'TextureProvider' = texture_provider
        """
        *For internal use only*

        The `TextureProvider` instance that allows us to
        have the functionality related to interacting
        with the texture pool.
        """

    def acquire(
        self,
        width: int = 1920,
        height: int = 1080,
        number_of_components: int = 4,
        dtype: str = 'f1'
    ) -> 'moderngl.Texture':
        """
        Get a `moderngl.Texture` instance with the given `width`,
        `height`, `number_of_components` and `dtype` from the
        pool.
        """
        return self._texture_provider.acquire(
            width = width,
            height = height,
            number_of_components = number_of_components,
            dtype = dtype
        )
    
    def acquire_like(
        self,
        texture: 'moderngl.Texture'
    ) -> 'moderngl.Texture':
        """
        Get the `moderngl.Texture` instance with the same
        properties than the `texture` provided as
        parameter.
        """
        return self._texture_provider.acquire_like(
            texture = texture
        )

    def release(
        self,
        texture: 'moderngl.Texture'
    ) -> None:
        """
        Release the `texture` provided, making it
        available again through the pool.
        """
        self._texture_provider.release(texture)


"""
The module to include classes related to the
rendering context.
"""
from yta_validation.parameter import ParameterValidator


class RenderContext:
    """
    A context that will be used along a rendering
    process, to be able to access to different
    resources, providers, etc.
    """

    def __init__(
        self,
        # TODO: Maybe rename to 'config'
        render_config: 'RenderConfig',
        # TODO: Maybe rename to 'resources'
        # TODO: Maybe rename to 'render_resources_handler'
        render_resources: 'RenderResources'
    ):
        ParameterValidator.validate_mandatory_instance_of('render_config', render_config, 'RenderConfig')
        ParameterValidator.validate_mandatory_instance_of('render_resources', render_resources, 'RenderResources')

        self.render_config: 'RenderConfig' = render_config
        """
        The configuration of the rendering process.
        """
        self.render_resources: 'RenderResources' = render_resources
        """
        The handler of the rendering process resources.
        """

    def deactivate_gpu(
        self
    ) -> 'RenderContext':
        """
        Deactivate the GPU by setting the `do_use_gpu`
        parameter to `False`, if needed.
        """
        self.render_resources.deactivate_gpu()

        return self

    def activate_gpu(
        self
    ) -> 'RenderContext':
        """
        Activate the GPU by setting the `do_use_gpu`
        parameter to `True`, if needed.
        """
        self.render_resources.activate_gpu()

        return self